import numpy as np
import numpy.lib.recfunctions as rfn
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d
import os
import scipy.stats
from scipy.stats import chi2
import sys
import subprocess
import pdb
import random as rd
import scipy as sp
import scipy.optimize as opti
import iminuit
from iminuit import Minuit



c_cmpers = 3.e10
n_percm3 =1.
h_kpc = 0.1
L_kpc = 5.
siginelH_mb = 30
siginelC_mb = 250
siginelB_mb = siginelC_mb
siginelBe_mb = siginelC_mb
siginelFe_mb = 700
sigprodCtoB_mb = 100.
D0_kpc2perMyr = 0.02
delta = 0.5
Vg_kmpers = 5.
Va_kmpers = 20
thalf10Be_Myr = 1.387
thalf14C_Myr = 7.73e-3
thalf26Al_Myr = 0.91
thalf36Cl_Myr = 0.307

mb_to_cm2 = 1.e-27
amu_to_GeV = 0.931
pc_to_m = 3.e16
s_to_Myr = 3.e-14
kmpers_to_kpcperMyr = 1.e-3
cm2pers_to_kpc2perMyr = 3e-30

#______________________________________________________________________________
def boron_to_carbon(R_GV,D0_kpc2perMyr=D0_kpc2perMyr,delta=delta,Vg_kmpers=Vg_kmpers,sinelB_mb=siginelB_mb,sprodCB_mb=sigprodCtoB_mb):
    beta = R2beta(10,5,R_GV)
    v = beta * c_cmpers / s_to_Myr
    nhv = n_percm3 * v * h_kpc
    nvh_inel = nhv * sinelB_mb * mb_to_cm2
    nvh_prod = nhv * sprodCB_mb * mb_to_cm2
    Vg = Vg_kmpers * kmpers_to_kpcperMyr
    D = beta * D0_kpc2perMyr * R_GV**delta
    if (Vg<1.e-5):
        return nvh_prod/(D/L_kpc+nvh_inel)
    else:
        return nvh_prod/(Vg / (1.-np.exp(-Vg*L_kpc/D)) + nvh_inel)

#______________________________________________________________________________
def be10to9_nowind(Ekn, D0_kpc2perMyr, L_kpc):
    R_GV = Ekn2R(10, 4,Ekn)
    beta = R2beta(10,4,R_GV)
    gamma = R2gamma(10,4,R_GV)
    v = beta * c_cmpers / s_to_Myr
    nhv = n_percm3 * v * h_kpc
    nvh_inel = nhv * siginelB_mb * mb_to_cm2
    D = beta * D0_kpc2perMyr * R_GV**delta
    l_rad = np.sqrt(gamma*thalf10Be_Myr/np.log(2.)*D)
    num = (D/L_kpc+nvh_inel)
    denom = (D/(l_rad*np.tanh(L_kpc/l_rad))+nvh_inel)
    return num/denom


#______________________________________________________________________________
def relcov_bc(l_corr):
    c=np.eye(n,n)
    for i in range (0,n):
        for j in range (0,n):
            c[i][j] = F[i]*F[j]*syst**2*np.exp(-0.5*(np.log(LRS[i]/LRS[j])/l)**2)
            if i==j:
                c[i][i] += (sig[i]/10)**2

#______________________________________________________________________________
def corr(cov):
    corr = cov.copy()
    n = len(cov[0])
    for i in range(n):
        for j in range(n):
            corr[i][j] = cov[i][j]/np.sqrt(cov[i][i]*cov[j][j])
    return corr

#______________________________________________________________________________
def err_stat(R):
    return 0.5*R**0.5

#______________________________________________________________________________
def err_syst(R):
    return 2*R**0.2


#        cova1=cov(l,syst).copy()
#        cova2=cov(1,syst).copy()
#        sigt=(np.linalg.eig(cova2)[0])**0.5
#        U=np.linalg.eig(cova2)[1]
#        for i in range (0,n):
#            y.append(boverc6(LRS[i],K0,delta,Vg,sinelB,sprodCB,hn))
#        yt = np.dot(np.linalg.inv(U),y)
#        for i in range (0,n):
#            g.append(rd.gauss(0,sigt[i]))
#            yt[i]+=g[i]
#        BCS=np.dot(U,yt)
#        x0=[K0,delta,Vg]
#        opt = opti.curve_fit(boverc3, LRS, BCS, x0)
#        deltcov=[]



#______________________________________________________________________________
def generate_BCdata(Rvals, l_corr=-1, is_siginel_bias='false'):
    # Always have data (sometimes biased)
    if is_siginel_bias=='false':
        # No bias
        data = boron_to_carbon(Rvals)
    else:
        # With 20% bias on inelastic XS
        data = boron_to_carbon(Rvals,sinelB_mb=(1.+0.1)*siginelB_mb)

    sigma_data = err_stat(Rvals)
    new_data = []
    err_data = []
    for i in range(0,len(Rvals)):
       err_data.append(sigma_data[i]/100.*data[i])
       new_data.append(rd.gauss(data[i],err_data[i]))

    # Always have statistical uncertainties
    #err_stat = extra.stat_err(Rvals)
    return new_data, err_data

#______________________________________________________________________________
def load_data(main_dir, filename):
    file = main_dir + '/' + filename
    data = np.genfromtxt(file, comments='#', usecols=(3,4,5,6,7,8,9,10),\
                         dtype='float, float, float, float, float, float, float, float',\
                         names=['R', 'RL', 'RU', 'dfdR', 'StatL', 'StatU', 'SystL', 'SystU'])

    # Update errors and store them as extra column in ndarray
    err_tot_L = np.sqrt(data['StatL']*data['StatL'] + data['SystL']*data['SystL'])
    err_tot_U = np.sqrt(data['StatU']*data['StatU'] + data['SystU']*data['SystU'])
    R_L = np.abs(data['R']-data['RL'])
    R_U = np.abs(data['R']-data['RU'])

    data_new = rfn.append_fields(data, names=['ErrTotL', 'ErrTotU', 'ErrRL', 'ErrRU'],\
                                 data=[err_tot_L,err_tot_U, R_L, R_U], usemask=False)
    return data_new

#______________________________________________________________________________
def load_cov(main_dir, file_name):
    data_cov = {}
    is_start_err = False
    with open(main_dir+'/'+file_name) as file:
        for line in file:
            # Skip comment
            if line.startswith('#'):
                continue
            # Add error type
            elif line.startswith('ErrorType:'):
                is_start_err = True
                err_type = line.split()[1]
            # Add as new line
            else:
                if is_start_err:
                    data_cov[err_type] = [line.split()]
                    is_start_err = False
                    continue
                else:
                    data_cov[err_type].append(line.split())
    # Also add matrix of total covariance
    l_keys = list(data_cov.keys())
    for i,key in enumerate(l_keys):
        if i == 0:
            data_cov['tot'] = data_cov[key].copy()
        else:
            data_cov['tot'] = np.add(np.asarray(data_cov['tot'],dtype=float), np.asarray(data_cov[key],dtype=float))

        #    for r in range(len(data_cov[key])):
        #        for c in range(len(data_cov[key])):
        #            data_cov['tot'][r][c] += data_cov[key][r][c]


    data_cov['tot_diag']= data_cov['tot'].copy()
    N_data = len(data_cov['tot'])
    for i in range(N_data):
            for j in range(N_data):
                if i!=j:
                    data_cov['tot_diag'][i][j] = 0.

    return data_cov


#______________________________________________________________________________
def R2beta(A,Z,R):
   m=A*amu_to_GeV
   return (R*Z)/(np.sqrt((R*Z)**2+m**2))

#______________________________________________________________________________
def R2gamma(A,Z,R):
   m=A*amu_to_GeV
   return np.sqrt((R*Z)**2+m**2)/m

#______________________________________________________________________________
def R2Ekn(A,Z,R, is_mequalsA):
   if (is_mequalsA):
      m=A*amu_to_GeV
   else:
      m=A
   return (np.sqrt((R*Z)**2+m**2)-m)/A

#______________________________________________________________________________
def Ekn2R(A,Z,Ekn):
   m=A*amu_to_GeV
   Etot = (Ekn*A+m)
   return np.sqrt(Etot*Etot-m*m)/Z

#______________________________________________________________________________
def sol_tdep_diff(n0,x,D,t):
   return n0/(np.sqrt(4.*np.pi*D*t))*np.exp(-x**2/(4*D*t))

#______________________________________________________________________________
def plot_bc():
   nR = 50
   Rvals = np.logspace(-1, 4., nR)
   plt.ion()
   plt.rcParams.update({'font.size': 16, 'legend.fontsize': 16, 'legend.handlelength': 2})
   plt.figure(figsize=(10,6))
   plt.title("Impact of transport parameters")
   plt.xlabel('Rigidity [GV]')
   plt.ylabel('B/C')
   plt.loglog()
   plt.plot(Rvals, boron_to_carbon(Rvals,D0_kpc2perMyr,delta,Vg_kmpers,siginelB_mb,sigprodCtoB_mb),label="Ref")
   plt.plot(Rvals, boron_to_carbon(Rvals,D0_kpc2perMyr,delta,Vg_kmpers,siginelB_mb,sigprodCtoB_mb/2.),label=r'$\sigma_{prod}^{\rm BC}/2$',linestyle = '--')
   plt.plot(Rvals, boron_to_carbon(Rvals,2.*D0_kpc2perMyr,delta,Vg_kmpers,siginelB_mb,sigprodCtoB_mb),label=r'$2 D_0$',linestyle = ':')
   plt.plot(Rvals, boron_to_carbon(Rvals,D0_kpc2perMyr,delta,Vg_kmpers,0.,sigprodCtoB_mb),label=r'$\sigma_{inel}^{\rm B} = 0$',linestyle = '-.')
   plt.plot(Rvals, boron_to_carbon(Rvals,D0_kpc2perMyr,delta,0.,siginelB_mb,sigprodCtoB_mb),label=r'$V_g=0$',linestyle = '--')

   data = load_data('.','IMAGES/BC.USINE')
   plt.errorbar(data['R'], data['dfdR'], yerr=[data['ErrTotL'],data['ErrTotU']], fmt='o', color='#000000', ms=1.5, elinewidth=0.5, capsize=1.0, label = 'AMS-02', markeredgewidth=0.5)

   plt.legend(bbox_to_anchor=(0.987, 0.982), loc=1, borderaxespad=0.)

#______________________________________________________________________________
def plot_be10to9_nowind():
   nEkn = 50
   Eknvals = np.logspace(-2, 3., nEkn)
   plt.ion()
   plt.rcParams.update({'font.size': 16, 'legend.fontsize': 16, 'legend.handlelength': 2})
   plt.figure(figsize=(10,6))
   plt.title(r"Lifting degeneracy with radioactive secondary")
   plt.xlabel(r'$E_{k/n}$ [GeV/n]')
   plt.ylabel(r'$^{10}{\rm Be}/^9{\rm Be}$')
   plt.semilogx()
   plt.plot(Eknvals, be10to9_nowind(Eknvals, D0_kpc2perMyr/2.5, L_kpc/2.5),label=r"$D_0/L=0.04$, $L=2$ kpc", linestyle = '--')
   plt.plot(Eknvals, be10to9_nowind(Eknvals, D0_kpc2perMyr, L_kpc),label=r"$D_0/L=0.04$, $L=5$kpc")
   plt.plot(Eknvals, be10to9_nowind(Eknvals, D0_kpc2perMyr*2, L_kpc*2),label=r"$D_0/L=0.04$, $L=10$ kpc", linestyle = ':')

   data = load_data('.','IMAGES/10Be9Be.USINE')
   plt.errorbar(data['R'], data['dfdR'], yerr=[data['ErrTotL'],data['ErrTotU']], fmt='o', color='#000000', ms=1.5, elinewidth=0.5, capsize=1.0, label = 'Data', markeredgewidth=0.5)

   plt.legend(bbox_to_anchor=(0.5, 0.982), loc=1, borderaxespad=0.)


#______________________________________________________________________________
def plot_tdep_diff():

   n0=1.
   D=0.01 #kpc^2/Myr
   x = np.arange(-10,10,0.1) #kpc

   plt.ion()
   plt.rcParams.update({'font.size': 16, 'legend.fontsize': 16, 'legend.handlelength': 2})
   plt.figure(figsize=(8,6))
   plot_times =  np.logspace(0,3,dtype='double', num=4) #Myr
   for t in plot_times:
      n_diff = sol_tdep_diff(n0, x, D, t)
      leg = 't=%.1le [Myr]'%(t)
      plt.plot(x, n_diff, linewidth=2, label=leg)
   plt.legend()
   plt.xlabel('Distance [Myr]',fontsize=12)
   plt.ylabel('$n(x,t) / n_0$',fontsize=12)
   plt.axis([-10,10,0,1.2*n0])
   plt.title('D=%.2f'%(D) + ' kpc$^2$ Myr$^{-1}$');
   #plt.show()


#______________________________________________________________________________
def plot_timescales():
    nEkn = 50
    Eknvals = np.logspace(-2, 4., nEkn)
    gamma_vals = Eknvals + 1.
    beta_vals =  np.sqrt(Eknvals * (Eknvals + 2.)) / (Eknvals+1.);
    #print(Eknvals, beta_vals, gamma_vals)


    t_conv = [L_kpc / (Vg_kmpers*kmpers_to_kpcperMyr)] * nEkn


    plt.ion()
    plt.rcParams.update({'font.size': 16, 'legend.fontsize': 16, 'legend.handlelength': 2})

    fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(16,10))

    # Nuclei
    Diff_kpcperMyr = D0_kpc2perMyr * beta_vals * Ekn2R(10,5,Eknvals)**delta
    t_diff = L_kpc**2 / (2. * Diff_kpcperMyr)
    t_decay_10Be = gamma_vals*thalf10Be_Myr/np.log(2.)
    t_decay_14C = gamma_vals*thalf14C_Myr/np.log(2.)
    t_decay_26Al = gamma_vals*thalf26Al_Myr/np.log(2.)
    t_decay_36Cl = gamma_vals*thalf36Cl_Myr/np.log(2.)
    t_dest_H = 1./(n_percm3 * beta_vals *(c_cmpers/s_to_Myr) * (siginelH_mb*mb_to_cm2) * (h_kpc/L_kpc))
    t_dest_C = 1./(n_percm3 * beta_vals *(c_cmpers/s_to_Myr) * (siginelC_mb*mb_to_cm2) * (h_kpc/L_kpc))
    t_dest_Fe = 1./(n_percm3 * beta_vals *(c_cmpers/s_to_Myr) * (siginelFe_mb*mb_to_cm2) * (h_kpc/L_kpc))
    t_reacc = (9.* Diff_kpcperMyr) / (2. * (Va_kmpers*kmpers_to_kpcperMyr*beta_vals)**2)
    t_ion_C = 300. * beta_vals**3 * (Eknvals*12 + 6*amu_to_GeV)
    ax[0].set_title("Nuclei")
    ax[0].set_xlabel(r'$E_{k/n}$ [GeV/n]')
    ax[0].set_ylabel('Time [Myr]')
    ax[0].loglog()
    ax[0].plot(Eknvals, t_diff, linewidth=2, label="Diff")
    ax[0].plot(Eknvals, t_conv, linewidth=2, label="Conv")
    ax[0].plot(Eknvals, t_reacc, linewidth=2, label="Reacc")
    ax[0].plot(Eknvals, t_ion_C, linewidth=2, label="Ion+Coul")
    ax[0].plot(Eknvals, t_decay_10Be, lw=3,  alpha=0.8, color='indigo', ls='--', label=r'$^{10}$Be')
    ax[0].plot(Eknvals, t_decay_26Al, lw=3,  alpha=0.6, color='indigo', ls='--', label=r'$^{26}$Al')
    ax[0].plot(Eknvals, t_decay_36Cl, lw=3,  alpha=0.4, color='indigo', ls='--', label=r'$^{36}$Cl')
    ax[0].plot(Eknvals, t_decay_14C, lw=3,  alpha=0.3, color='indigo', ls='--', label=r'$^{14}$C')
    ax[0].plot(Eknvals, t_dest_H, lw=4, alpha=0.1, color='black', ls='-.', label=r'$\sigma_{\rm inel}$(H)')
    ax[0].plot(Eknvals, t_dest_C, lw=4, alpha=0.2, color='black', ls='-.', label=r'$\sigma_{\rm inel}$(C)')
    ax[0].plot(Eknvals, t_dest_Fe, lw=4, alpha=0.3,color='black', ls='-.', label=r'$\sigma_{\rm inel}$(Fe)')
    ax[0].set_ylim([1,5.e5])
    ax[0].legend(bbox_to_anchor=(0.987, 0.982), loc=1, borderaxespad=0., ncol=3)

    # Leptons
    nEk = 50
    Ekvals = np.logspace(-2, 4., nEk)
    Etotvals = Ekvals + 511.e-6
    pvals = np.sqrt(Etotvals*Etotvals-511.e-6*511.e-6)
    beta_vals = pvals/Etotvals
    gamma_vals = Etotvals/511.e-6
    Diff_kpcperMyr = D0_kpc2perMyr * beta_vals * pvals**delta
    t_reacc_leptons = (9.* Diff_kpcperMyr) / (2. * (Va_kmpers*kmpers_to_kpcperMyr)**2)
    t_ion_leptons = 300. * beta_vals**3 * Etotvals
    t_IC = 300. / Etotvals
    t_brem = [300.*(L_kpc/h_kpc)] * nEk
    ax[1].set_title("Leptons")
    ax[1].set_xlabel(r'$E_k$ [GeV/n]')
    ax[1].set_ylabel('Time [Myr]')
    ax[1].loglog()
    ax[1].plot(Ekvals, t_diff, linewidth=2, label="Diff")
    ax[1].plot(Ekvals, t_conv, linewidth=2, label="Conv")
    ax[1].plot(Ekvals, t_reacc, linewidth=2, label="Reacc")
    ax[1].plot(Ekvals, t_ion_C, linewidth=2, label="Ion+Coulomb")
    ax[1].plot(Ekvals, t_IC, lw=3, ls='--', label=r'IC+synchrotron')
    ax[1].plot(Ekvals, t_brem, lw=3, ls=':', label=r'Bremsstrahlung')
    ax[1].set_ylim([1,5.e5])
    ax[1].legend(bbox_to_anchor=(0.987, 0.982), loc=1, borderaxespad=0., ncol=2)
